package org.codehaus.plexus.util.io;

/*
 * Copyright The Codehaus Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.IOException;
import java.io.InputStream;

/**
 * Implementation of {@link InputStreamFacade} for raw input streams.
 */
@SuppressWarnings( { "UnusedDeclaration" } )
public class RawInputStreamFacade implements InputStreamFacade {
    final InputStream stream;

    public RawInputStreamFacade( InputStream stream )
    {
        this.stream = stream;
    }

    public InputStream getInputStream() throws IOException {
        return stream;
    }
}
